import 'dart:ui';

import '../extensions/extensions.dart';
import 'package:meta/meta.dart';

import '../utils/order_list.dart';

abstract class Renderer {
  double _priority;

  bool _enableTickToParent = true;
  bool get enableTickToParent => _enableTickToParent;

  Renderer({double? priority}) : _priority = priority ?? 0;

  final List<Renderer> _children = [];
  bool get hasChildren => _children.isNotEmpty;
  List<Renderer> get children => _children;

  Renderer? _parent;
  Renderer? get parent => _parent;

  double get priority => _priority;
  set priority(double next) {
    if (_priority == next) return;
    final prevPriority = _priority;

    _priority = next;
    _parent?.rebalance(this);
    onPriorityChanged(prevPriority, _priority);
  }

  void onPriorityChanged(double prev, double next) {}

  void rebalance(Renderer child) {
    if (_children.remove(child)) {
      orderList(_children, child, _getCompareValue);
      child.tick();
    }
  }

  void add(Renderer child) {
    assert(child._parent == null);
    orderList(_children, child, _getCompareValue);
    child
      .._parent = this
      ..tick()
      ..onMount();
  }

  void addAll(Iterable<Renderer> children) {
    children.forEach(add);
  }

  void onMount() {}

  void remove(Renderer child) {
    assert(child._parent != null);
    _children.remove(child);
    child
      ..tick()
      .._parent = null
      ..onRemove();
  }

  void removeAll(Iterable<Renderer> children) {
    children.forEach(remove);
  }

  void removeFromParent() {
    _parent?.remove(this);
  }

  void onRemove() {}

  @mustCallSuper
  void onCanvasResize(Vector2 canvasSize) {
    for (int i = _children.length; i-- > 0;) {
      _children[i].onCanvasResize(canvasSize);
    }
  }

  bool propagateToChildren<T extends Renderer>(
    bool Function(T child) handler, {
    bool includeSelf = false,
  }) {
    if (includeSelf && this is T && !handler(this as T)) {
      return false;
    }

    for (int i = 0; i < _children.length; ++i) {
      if (!_children[i].propagateToChildren<T>(handler, includeSelf: true)) {
        return false;
      }
    }

    return true;
  }

  T? find<T extends Renderer>(
    bool Function(T child) handler, {
    bool includeSelf = false,
  }) {
    T? target;

    propagateToChildren<T>((child) {
      if (handler(child)) {
        target = child;
        return false;
      }
      return true;
    }, includeSelf: includeSelf);

    return target;
  }

  Iterable<T> ancestors<T extends Renderer>({bool includeSelf = false}) sync* {
    var current = includeSelf && this is T ? this : _parent;
    while (current != null) {
      if (current is T) {
        yield current;
      }
      current = current._parent;
    }
  }

  @mustCallSuper
  void tick({bool fromChild = false}) {
    if (_enableTickToParent) {
      _parent?.tick(fromChild: true);
    }
  }

  void disableTick(VoidCallback when) {
    _enableTickToParent = false;
    when();
    _enableTickToParent = true;
  }

  void renderWrapper(Canvas canvas) => render(canvas);

  void render(Canvas canvas) {}

  bool beforeRender() {
    return true;
  }

  @mustCallSuper
  bool beforeRenderChildren() {
    return hasChildren;
  }

  void renderChildrenWrapper(Canvas canvas) => renderChildren(canvas);

  void renderChildren(Canvas canvas) {
    for (int i = _children.length; i-- > 0;) {
      _children[i].renderTree(canvas);
    }
  }

  void renderTree(Canvas canvas) {
    final needRender = beforeRender();
    final needRenderChildren = beforeRenderChildren();

    if (!needRender && !needRenderChildren) return;

    transformCanvas(canvas, () {
      if (needRender) renderWrapper(canvas);
      if (needRenderChildren) renderChildrenWrapper(canvas);
    });
  }

  void updateTree() {
    update();
    updateChildren();
  }

  void update() {}

  void updateChildren() {
    for (int i = 0; i < _children.length; ++i) {
      _children[i].updateTree();
    }
  }

  void transformCanvas(Canvas canvas, VoidCallback callback) {
    callback();
  }

  bool containsPoint(Vector2 point) => false;

  static num _getCompareValue(Renderer child) {
    return child.priority;
  }
}
